/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.druid.indexer.partitions;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;

import javax.annotation.Nullable;
import java.util.Objects;

/**
 * Dynamically determine partitions in the middle of indexing.
 */
public class DynamicPartitionsSpec implements PartitionsSpec
{
  public static final long DEFAULT_MAX_TOTAL_ROWS = 20_000_000;

  private final int maxRowsPerSegment;
  private final long maxTotalRows;

  @JsonCreator
  public DynamicPartitionsSpec(
      @JsonProperty("maxRowsPerSegment") @Nullable Integer maxRowsPerSegment,
      @JsonProperty("maxTotalRows") @Nullable Long maxTotalRows
  )
  {
    this.maxRowsPerSegment = PartitionsSpec.isEffectivelyNull(maxRowsPerSegment)
                             ? DEFAULT_MAX_ROWS_PER_SEGMENT
                             : maxRowsPerSegment;
    this.maxTotalRows = PartitionsSpec.isEffectivelyNull(maxTotalRows) ? DEFAULT_MAX_TOTAL_ROWS : maxTotalRows;
  }

  @Override
  @JsonProperty
  public Integer getMaxRowsPerSegment()
  {
    return maxRowsPerSegment;
  }

  @JsonProperty
  public long getMaxTotalRows()
  {
    return maxTotalRows;
  }

  @Override
  public boolean needsDeterminePartitions(boolean useForHadoopTask)
  {
    return false;
  }

  @Override
  public boolean equals(Object o)
  {
    if (this == o) {
      return true;
    }
    if (o == null || getClass() != o.getClass()) {
      return false;
    }
    DynamicPartitionsSpec that = (DynamicPartitionsSpec) o;
    return maxRowsPerSegment == that.maxRowsPerSegment &&
           maxTotalRows == that.maxTotalRows;
  }

  @Override
  public int hashCode()
  {
    return Objects.hash(maxRowsPerSegment, maxTotalRows);
  }

  @Override
  public String toString()
  {
    return "DynamicPartitionsSpec{" +
           "maxRowsPerSegment=" + maxRowsPerSegment +
           ", maxTotalRows=" + maxTotalRows +
           '}';
  }
}
